#!/usr/bin/env python

"""Test if there is TCP censorship between two non-measurement machines."""

__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

from collections import deque
from math import ceil
import sys

from dpkt.tcp import TH_SYN, TH_ACK, TCPFlags

from intervention import intervention_at_time
from nslookup4 import lookup_or_die
from scanerror import ScanError
from sequence import TCPSequence
from sniffing import start_sniffing
from synack4 import synack_retransmit_times
from tcp4 import receive_tcp
from tcpratetest4 import tcp_rate_test
from tcptest4 import tcp_detect_ip_id_pattern

ALPHA = 0.01
INTERVENTION = 5
SAMPLES = 100
SLEEP_SMALL = 0.1
TIMEOUT = 1.0
WINDOW = 20

def test_censor_case(sniffer, saddr, caddr, spoofed_addr, port):
    """Test if 'caddr' can connect to 'saddr' on port 'port' using
    'spoofed_addr' as a spoofed address for measuring SYN/ACK transmission
    times.

    This function will return True if there is censorship from the server to
    client, False is there is censorship from the client to the server, else
    None.

    Choose 'spoofed_addr' as an unused address on your subnet that is not bound
    to your (or anyone's) interface so that your kernel will not respond to the
    incoming SYN/ACK's with RST's.   This address will be ARP-spoofed so that
    responses will still be sniffed.  This is used for measuring saddr's
    SYN/ACK retransmission patterns (see synack4.retransmit_times).

    This test spoofs SYN's from caddr to saddr.  It then uses caddr's globally
    incrementing IP id as a side channel to determine if saddr's SYN/ACK's
    are reaching caddr and to see if caddr's RST's are reaching saddr.  If
    saddr's SYN/ACK's aren't reaching caddr, then caddr will not respond and
    there will be no effect.  If saddr's SYN/ACK's are reaching caddr, and
    caddr's RST's are reaching saddr, then caddr's global IP id counter will
    increase by one.  If caddr's RST's are not reaching saddr, saddr will
    retransmit multiple SYN/ACK's and caddr will respond to each one with a
    RST, so caddr's global IP id counter will increase by over one.

    Before this test, caddr will first be tested to ensure that it has a
    globally incrementing IP ID and that it transmits enough RST's in response
    to SYN/ACK's.  saddr will also be tested to ensure that it retrasmits
    enough SYN/ACK's in response to SYN's.
    """
    sys.stderr.write('Testing client\'s IP id pattern...\n')
    flags = TCPFlags(TH_SYN | TH_ACK)
    pattern = tcp_detect_ip_id_pattern(sniffer, [caddr], spoofed_addr, sport=port, flags=flags, verbose=True)[caddr]
    if pattern['type'] != 'globally':
        raise ScanError('Client has \'%s\' IP id pattern, not globally incrementing' % pattern['type'])
    sys.stderr.write('Testing server\'s SYN/ACK retransmission pattern...\n')
    retransmit_times, stats = synack_retransmit_times(sniffer, saddr, spoofed_addr, port, window=WINDOW, repeat=INTERVENTION, verbose=True)
    sys.stderr.write('Retransmit times: %s\n' % retransmit_times)
    if not retransmit_times:
        raise ScanError('Cannot get SYNACK retransmit times from server on port %d' % port)
    if len(retransmit_times) < 2:
        raise ScanError('Server on port %d must transmit at least two SYNACK\'s' % port)
    k1 = 0.5 * INTERVENTION
    k2 = min(2, (len(retransmit_times) + 1.0) / 2.0) * INTERVENTION
    if stats['syn_frac'] <= 0.5 or stats['avg_trans'] * INTERVENTION <= k2:
        raise ScanError('Server on port %d responds with too few SYNACK transmissions' % port)
    sys.stderr.write('Testing client for RST rate-limiting...\n')
    rate = tcp_rate_test(sniffer, caddr, len(retransmit_times) * INTERVENTION, sport=port, flags=flags, verbose=True)
    if rate['avg'] <= k2:
        raise ScanError('Client responds with too few RST\'s')
    retransmit_times = [int(ceil(t)) for t in retransmit_times]
    retransmit_times[0] = 0
    reset_delay = retransmit_times[-1] + 1
    ts = [t + SAMPLES // 2 for t in retransmit_times]
    last_ip_id = None
    xs = []
    rsts_queue = deque()
    seq = TCPSequence(caddr, iface=sniffer.iface)
    seq_spoofed = TCPSequence(saddr, caddr)
    timeout = sniffer.timeout()
    for i in xrange(SAMPLES + 1):
        if i > SAMPLES // 2:
            sniffer.send(requests)
            for request in requests:
                seq_spoofed.rst(request)
            rsts_queue.append(requests)
        timeout.sleep(SLEEP_SMALL)
        x = seq.next_packet(sport=port, flags=flags)
        sniffer.send(x)
        if i > SAMPLES // 2 + reset_delay:
            rsts = rsts_queue.popleft()
            sniffer.send(rsts)
        if SAMPLES // 2 <= i < SAMPLES:
            # generate next packets (expensive) before next loop
            requests = [seq_spoofed.next_packet(dport=port) for _ in xrange(INTERVENTION)]
        sniffer.heartbeat()
        timeout.add_time(TIMEOUT - SLEEP_SMALL)
        p = receive_tcp(sniffer, x, timeout)
        ip_id = None if p is None else p.id
        if i > 0:
            diff = None if ip_id is None or last_ip_id is None else ip_id - last_ip_id - 1 & 0xffff
            sys.stderr.write('%s%s: %s\n' % (ip_id, '*' * sum(i > t for t in ts), diff))
            xs.append(diff)
        last_ip_id = ip_id
        timeout.sleep()
    while rsts_queue:
        rsts = rsts_queue.popleft()
        sniffer.send(rsts)
    try:
        ir = intervention_at_time(xs, ts, [k1, k2], 0.05, True)
    except Exception:
        raise ScanError('Client traffic is too noisy or lossy--try another client?')
    sys.stderr.write('%s\n' % ir)
    if ir.pvalues[0] < ALPHA:
        return True
    elif ir.pvalues[0] > 1.0 - ALPHA:
        if ir.pvalues[1] < ALPHA:
            return None
        elif ir.pvalues[1] > 1.0 - ALPHA:
            return False
    raise ScanError('Client traffic is too noisy or lossy--try another client?')

def usage(code=1):
    sys.stderr.write('Usage: %s SERVER PORT CLIENT SPOOFED_HOST\n' % sys.argv[0])
    exit(code)

def main():
    try:
        if len(sys.argv) != 5:
            usage()
        shost = sys.argv[1]
        try:
            port = int(sys.argv[2])
        except ValueError:
            usage()
        chost = sys.argv[3]
        spoofed_host = sys.argv[4]
        saddr = lookup_or_die(shost)[0]
        caddr = lookup_or_die(chost)[0]
        spoofed_addr = lookup_or_die(spoofed_host)[0]
        is_server_to_client = start_sniffing(test_censor_case, False, args=(saddr, caddr, spoofed_addr, port))
        if is_server_to_client is None:
            print 'No censorship'
        elif is_server_to_client:
            print 'Server to client censorship'
        else:
            print 'Client to server censorship'
    except ScanError as e:
        sys.stderr.write('Error performing scan: %s\n' % e)
    except KeyboardInterrupt:
        sys.stderr.write('Received SIGINT\n')

if __name__ == '__main__':
    main()
