#!/usr/bin/env python

"""Receive ICMPv6 echo replies answering echo requests."""

__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

import socket
import sys

from dpkt.dpkt import Packet
from dpkt.icmp6 import ICMP6, ICMP6_ECHO_REPLY

from nslookup6 import lookup_or_die
from sequence import ICMP6Sequence
from sniffing import start_sniffing

def receive_icmps(sniffer, sents, timeout, fragments=False):
    """Return a generator generating ICMP packets answering any packets in
    'sents' until 'timeout' has expired.

    An ICMP packet y answers another x if x is an echo request, y is an echo
    reply, and they both have the same ICMP echo id's and sequence numbers.
    If 'fragments' is specified, incoming fragments with offsets > 0 will be
    matched based on ICMP data.
    """
    if isinstance(sents, Packet):
        sents = [sents]
    sents_set = set()
    for sent in sents:
        sent_icmp = sent.data
        sent_echo = sent_icmp.data
        key = sent.dst, sent_echo.id, sent_echo.seq
        sents_set.add(key)
    while True:
        p = sniffer.sniff(timeout)
        if p is None:
            break
        p_frag = p.extension_hdrs[44]
        if p_frag is None or p_frag.frag_off_resv_m & 0xfff8 == 0:
            if type(p.data) is ICMP6 and p.data.type == ICMP6_ECHO_REPLY:
                p_icmp = p.data
                p_echo = p_icmp.data
                key = p.src, p_echo.id, p_echo.seq
                if key in sents_set:
                    yield p
        elif fragments and p.p == socket.IPPROTO_ICMPV6:
            p_load = str(p.data)[:8]
            for sent in sents:
                # only check ICMP data
                sent_load = str(sent_echo.data)
                sent_load = sent_load[-(((len(sent_load) - 1) % 8) + 1):]
                # assume that data is 8-byte periodic
                if p_load == sent_load:
                    yield p

def receive_icmp(*args, **kwargs):
    """Return the first ICMP packet answering 'sents', or None if 'timeout'
    has expired.

    See receive_icmps for more details.
    """
    return next(receive_icmps(*args, **kwargs), None)

def ping(sniffer, addr, mtu):
    seq = ICMP6Sequence(addr, iface=sniffer.iface)
    timeout = sniffer.timeout()
    while True:
        timeout.add_time(1.0)
        request = seq.next_packet_testing_mtu(mtu)
        print 'Sending %d...' % request.extension_hdrs[44].id
        sniffer.send(request, True)
        p = None
        for p in receive_icmps(sniffer, request, timeout, True):
            frag = p.extension_hdrs[44]
            if frag is not None:
                print '%d+%d' % (frag.id, frag.frag_off_resv_m & 0xfff8)
            else:
                print '*+0'
        if p is None:
            print '*'
        sys.stdout.flush()

def usage(code=1):
    sys.stderr.write('Usage: %s HOST MTU\n' % sys.argv[0])
    exit(code)

def main():
    try:
        if len(sys.argv) != 3:
            usage()
        host = sys.argv[1]
        try:
            mtu = int(sys.argv[2])
        except ValueError:
            usage()
        addr = lookup_or_die(host)[0]
        start_sniffing(ping, True, args=(addr, mtu))
    except KeyboardInterrupt:
        sys.stderr.write('Received SIGINT\n')

if __name__ == '__main__':
    main()
