"""Packet sending and sniffing with associated utility functions."""

__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

from collections import deque
from contextlib import closing
from fcntl import ioctl
import os
from select import select
import signal
import socket
import struct
import threading

from dpkt.dpkt import Packet, in_cksum
from dpkt.arp import ARP
from dpkt.icmp6 import ICMP6, ND_NEIGHBOR_SOLICIT, ND_NEIGHBOR_ADVERT
from dpkt.ip import IP, IP_MF
from dpkt.ip6 import IP6, ext_hdrs

from timeout import Timeout

ETH_P_ALL = 0x0003
ETH_P_ARP = 0x0806
ETH_P_IP = 0x0800
ETH_P_IPV6 = 0x86dd

STRUCT16BE = struct.Struct('!H')

def fragmentbuf(p, fragsize=1480, firstsize=None):
    """Fragments an IPv4 datagram into a list of byte strings.

    All but the list's last packet payload have size fragsize (rounded up to
    the nearest multiple of 8) and the last packet's payload size is no
    greater.  The firstsize may be optionally specified to have a special-case
    payload size for the first fragment.
    """
    if type(p) is not IP:
        raise ValueError('p must be IP packet')
    if firstsize is None:
        firstsize = fragsize
    fragsize = (fragsize + 7) // 8 * 8
    firstsize = (firstsize + 7) // 8 * 8
    buf = str(p)
    hdrlen = len(buf) - len(p.data)
    hdr, s = buf[:hdrlen], buf[hdrlen:]
    ihl = hdrlen // 4
    off = p.off & ~0x1fff | IP_MF
    hdr = bytearray(hdr)
    nb = 1 + max((len(s) + fragsize - firstsize - 1) // fragsize, 0)
    lst = [None] * nb
    for i in xrange(nb):
        if i == nb - 1:
            off &= ~IP_MF
        offset = 0 if i == 0 else firstsize + fragsize * (i - 1)
        size = firstsize if i == 0 else fragsize
        load = s[offset:offset + size]
        STRUCT16BE.pack_into(hdr, 2, ihl * 4 + len(load))
        STRUCT16BE.pack_into(hdr, 6, off | (offset // 8 & 0x1fff))
        hdr[10] = hdr[11] = 0
        STRUCT16BE.pack_into(hdr, 10, in_cksum(str(hdr)))
        lst[i] = str(hdr) + load
    return lst

def fragmentbuf6(p, fragsize=1280, firstsize=None):
    """Fragments an IPv6 datagram into a list of byte strings.

    All but the list's last packet have size fragsize (rounded up to the
    nearest multiple of 8) and the last packet's size is no greater.  The
    firstsize may be optionally specified to have a special-case size for the
    first fragment.
    """
    if type(p) is not IP6:
        raise ValueError('p must be IP6 packet')
    if p.extension_hdrs[44] is None:
        raise ValueError('p must contain IP6FragmentHeader')
    frag = p.extension_hdrs[44]
    if firstsize is None:
        firstsize = fragsize
    fragsize = (fragsize + 7) // 8 * 8
    firstsize = (firstsize + 7) // 8 * 8
    buf = str(p)
    hdrlen = 40
    for ext_hdr_number in ext_hdrs:
        ext_hdr = p.extension_hdrs[ext_hdr_number]
        if ext_hdr is not None:
            hdrlen += len(ext_hdr)
        if ext_hdr_number == 44:
            break
    hdr, s = buf[:hdrlen], buf[hdrlen:]
    fragsize -= hdrlen
    firstsize -= hdrlen
    frag_off_resv_m = frag.frag_off_resv_m & ~0xfff8 | 1
    hdr = bytearray(hdr)
    nb = 1 + max((len(s) + fragsize - firstsize - 1) // fragsize, 0)
    lst = [None] * nb
    for i in xrange(nb):
        if i == nb - 1:
            frag_off_resv_m &= ~1
        offset = 0 if i == 0 else firstsize + fragsize * (i - 1)
        size = firstsize if i == 0 else fragsize
        load = s[offset:offset + size]
        STRUCT16BE.pack_into(hdr, 4, hdrlen - 40 + len(load))
        STRUCT16BE.pack_into(hdr, hdrlen - 6, frag_off_resv_m | offset)
        lst[i] = str(hdr) + load
    return lst

def iface_to_mac(iface):
    """Returns mac address of the given interface."""
    SIOCGIFHWADDR = 0x8927
    padded = iface + '\x00' * (32 - len(iface))
    with closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_IP)) as s:
        addr = ioctl(s, SIOCGIFHWADDR, padded)[18:24]
    return addr

def spoof4(buf, sa, addrs, s):
    """ARP-spoof IPv4 addresses.

    Examines 'buf' and AF_PACKET socket address 'sa' to see if an ARP request
    was received.  If so, and the requested address is in 'addrs', a spoofed
    ARP reply will be sent over AF_PACKET socket 's'.
    """
    iface, protocol, _, _, addr = sa
    if protocol == ETH_P_ARP:
        try:
            p = ARP(buf)
        except Exception:
            pass
        else:
            if p.op == 1 and p.tpa in addrs:
                sha = iface_to_mac(iface)
                reply = ARP(op=2, spa=p.tpa, tpa=p.spa, sha=sha, tha=p.sha)
                s.sendto(str(reply), (iface, protocol, 0, 0, addr))

def spoof6(buf, sa, addrs, s):
    """Neighbor-spoof IPv6 addresses.

    Examines 'buf' and AF_PACKET socket address 'sa' to see if an IPv6
    neighbor solicitation was received.  If so, and the requested address is in
    'addrs', a spoofed neighbor advertisement will be sent over AF_PACKET
    socket 's'.
    """
    iface, protocol, _, _, addr = sa
    if protocol == ETH_P_IPV6:
        try:
            p = IP6(buf)
        except Exception:
            pass
        else:
            if type(p.data) is ICMP6:
                icmp = p.data
                if icmp.type == ND_NEIGHBOR_SOLICIT:
                    neighbor = icmp.data
                    if neighbor.ta in addrs:
                        sha = iface_to_mac(iface)
                        reply = IP6(hlim=255, plen=32,
                                    nxt=socket.IPPROTO_ICMPV6,
                                    dst=p.src, src=neighbor.ta,
                                    p=socket.IPPROTO_ICMPV6, data=
                                    ICMP6(type=ND_NEIGHBOR_ADVERT, data=
                                    ICMP6.Neighbor(flags=0x60000000,
                                                   ta=neighbor.ta,
                                                   data='\x02\x01' + sha)))
                        s.sendto(str(reply), (iface, protocol, 0, 0, addr))

def _assert_main_thread():
    assert(threading.current_thread().name == 'MainThread')

def _try_f(f, *args, **kwargs):
    try:
        return f(*args, **kwargs)
    except Exception:
        return None

class InterruptedException(Exception):
    """Exception raised when one thread wants the current thread to prematurely
    terminate.
    """
    pass

class BlockingTimeout(Timeout):
    """A Timeout object that supports interruptable sleep.

    Each BlockingTimeout is associated with a threading.event object, one
    either passed in during construction or else one automatically created.
    Once the BlockingTimeout's event is set, then all present and future sleeps
    on that Timeout will be interrupted.
    """
    def __init__(self, delta=0.0, event=None):
        """Create Timeout object with timeout 'delta' seconds in the future.

        If an optional event is given, then that event is used to determine if
        this timeout instance has been interrupted.  Otherwise, an event will
        be automatically created for this Timeout object.
        """
        super(BlockingTimeout, self).__init__(delta=delta)
        self.event = threading.Event() if event is None else event
    def sleep(self, delta=0.0):
        """Sleeps until timeout has expired or the BlockingTimeout's event has
        been set, optionally adding 'delta' time until next timeout.
        """
        if self.event.wait(self.timeout(delta)):
            raise InterruptedException()

class Queue(object):
    """An interruptable, thread-safe blocking queue."""
    def __init__(self):
        self.condition = threading.Condition()
        self.queue = deque()
        self.interrupted = False
    def interrupt(self):
        """Interrupt all present and future gets on this Queue."""
        with self.condition:
            self.interrupted = True
            self.queue.clear()
            self.condition.notify_all()
    def put(self, item):
        """Put an item on this Queue."""
        with self.condition:
            if not self.interrupted:
                self.queue.append(item)
                self.condition.notify()
    def get(self, timeout):
        """Get next item on this Queue.

        If an item is not available, block until one is available or until this
        Queue is interrupted.
        """
        with self.condition:
            while True:
                if self.interrupted:
                    raise InterruptedException()
                if len(self.queue) > 0:
                    return self.queue.popleft()
                remaining = timeout.timeout()
                if remaining <= 0.0:
                    return None
                self.condition.wait(remaining)
    def clear(self):
        """Remove all items from Queue."""
        with self.condition:
            self.queue.clear()

def start_session(f):
    """Decorator for wrapping a function in a new Sniffer session."""
    def g(sniffer, *args, **kwargs):
        sniffer.start_session()
        try:
            return f(sniffer, *args, **kwargs)
        finally:
            sniffer.stop_session()
    return g

class Sniffer(object):
    """A thread-safe abstraction for sending and sniffing packets."""
    def __init__(self, ip6=False, subscribe=False, iface=None):
        self.timeout_event = threading.Event()
        self.local = threading.local()
        self.lock = threading.Lock()
        self.queues = {}
        self.queues_lock = threading.Lock()
        self.spoof_addrs = {}
        self.spoof_addrs_lock = threading.Lock()
        self.interrupted = False
        self.thread = None
        if ip6:
            self.domain = socket.AF_INET6
            self.type = ETH_P_IPV6
            self.socket_type = ETH_P_IPV6
            self.cls = IP6
            self.spoof = lambda buf, sa: spoof6(buf, sa, self.spoof_addrs, self.ins)
            self.fragment = fragmentbuf6
        else:
            self.domain = socket.AF_INET
            self.type = ETH_P_IP
            self.socket_type = ETH_P_ALL
            self.cls = IP
            self.spoof = lambda buf, sa: spoof4(buf, sa, self.spoof_addrs, self.ins)
            self.fragment = fragmentbuf
        self.iface = iface
        if subscribe:
            self._subscribe()
    def _sniff_socket(self):
        while True:
            rlist, wlist, xlist = select([self.ins, self.rpipe], [], [])
            if self.rpipe in rlist:
                break
            elif self.ins in rlist:
                buf, sa = self.ins.recvfrom(0xffff)
                self.spoof(buf, sa)
                iface, protocol, pkttype, _, addr = sa
                if pkttype == 0 and protocol == self.type:
                    with self.queues_lock:
                        for queue in self.queues.itervalues():
                            queue.put(buf)
    def _init(self):
        try:
            self.outs = socket.socket(self.domain, socket.SOCK_RAW, socket.IPPROTO_RAW)
            if self.iface is not None:
                SO_BINDTODEVICE = 25
                self.outs.setsockopt(socket.SOL_SOCKET, SO_BINDTODEVICE, self.iface + '\x00')
            self.outs.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 2 << 15)
            self.ins = socket.socket(socket.AF_PACKET, socket.SOCK_DGRAM, socket.htons(self.socket_type))
            if self.iface is not None:
                self.ins.bind((self.iface, self.socket_type))
            self.ins.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 2 << 15)
            self.rpipe, self.wpipe = os.pipe()
            self.thread = threading.Thread(target=self._sniff_socket)
            self.thread.start()
        except:
            self._uninit()
            raise
    def _uninit(self):
        if threading.current_thread() != self.thread:
            _try_f(lambda: os.write(self.wpipe, 'x'))
            _try_f(lambda: self.thread.join())
        _try_f(lambda: os.close(self.wpipe))
        _try_f(lambda: os.close(self.rpipe))
        _try_f(lambda: self.ins.close())
        _try_f(lambda: self.outs.close())
    def _subscribe(self):
        t = threading.current_thread()
        if t not in self.queues:
            self.local.callback = None
            self.local.sessions = []
            with self.lock:
                if self.interrupted:
                    raise InterruptedException()
                with self.queues_lock:
                    self.queues[t] = Queue()
                    init = len(self.queues) == 1
                if init:
                    self._init()
    def _unsubscribe(self):
        t = threading.current_thread()
        if t in self.queues:
            with self.lock:
                with self.queues_lock:
                    del self.queues[t]
                    uninit = len(self.queues) == 0
                if uninit:
                    self._uninit()
    def start_thread(self, target, args, kwargs):
        """Start a new thread subscribed to this Sniffer."""
        def wrapper():
            try:
                self._subscribe()
                target(self, *args, **kwargs)
            except InterruptedException:
                pass
            finally:
                self._unsubscribe()
        t = threading.Thread(target=wrapper)
        t.start()
        return t
    def __enter__(self):
        return self
    def __exit__(self, type, value, traceback):
        self.shutdown()
    def sniff(self, timeout):
        """Return next sniffed packet.

        If one is not available, block until one is available, the given
        timeout expires, or this Sniffer is shutdown.
        """
        queue = self.queues[threading.current_thread()]
        while True:
            buf = queue.get(timeout=timeout)
            if buf is None:
                return None
            try:
                return self.cls(buf)
            except Exception:
                pass
    def send(self, ps, fragment=False):
        """Send a packet or list of packets, optionally fragmenting them
        first.
        """
        if isinstance(ps, Packet):
            ps = [ps]
        for p in ps:
            self.sendbuf(self.fragment(p) if fragment else str(p), p.dst)
    def sendbuf(self, bufs, dst):
        """Send a byte string or list of byte strings."""
        if isinstance(bufs, str):
            bufs = [bufs]
        outs = self.outs
        for buf in bufs:
            outs.sendto(buf, (dst.ascii, 0))
    def shutdown(self):
        """Shut down this Sniffer.

        This will interrupt this Sniffer's Queue and all BlockingTimeout
        instances created through this Sniffer.
        """
        self._unsubscribe()
        with self.lock:
            self.interrupted = True
            self.timeout_event.set()
            with self.queues_lock:
                for queue in self.queues.itervalues():
                    queue.interrupt()
        _try_f(lambda: self.thread.join())
    def timeout(self, timeout=0.0):
        """Create a new BlockingTimeout sharing this Sniffer's event instance.

        These BlockingTimeout instances will be automatically interrupted when
        this Sniffer is shutdown.
        """
        return BlockingTimeout(timeout, self.timeout_event)
    def set_callback(self, callback):
        """Set this Sniffer's callback function.

        If not None, the callback function will be called when the heartbeat()
        method is invoked.  This can be used as a heartbeat to, e.g.,
        periodically refresh in a database that the scan is still running.
        """
        self.local.callback = callback
    def heartbeat(self):
        """Invoke this Sniffer's callback function.

        This should be called at least once a minute during long-running scans.
        """
        if self.local.callback is not None:
            self.local.callback()
    def start_session(self):
        """Start a new Sniffer session.

        All callback function changes and spoofed addresses will be rolled back
        when the session is ended.
        """
        self.local.sessions.append((self.local.callback, {}))
    def stop_session(self):
        """End a Sniffer session.

        All callback function changes and spoofed addresses since the last
        session was started will be rolled back.
        """
        try:
            callback, spoof_addrs = self.local.sessions.pop()
        except KeyError:
            pass
        else:
            self.local.callback = callback
            for addr, count in spoof_addrs.iteritems():
                if count >= 0:
                    for _ in xrange(count):
                        self._spoof_addr_remove(addr)
                else:
                    for _ in xrange(-count):
                        self._spoof_addr_add(addr)
    def _spoof_addr_add(self, n):
        with self.spoof_addrs_lock:
            count = self.spoof_addrs.get(n, 0) + 1
            self.spoof_addrs[n] = count
        return count
    def spoof_addr_add(self, addr):
        """Add an address to spoof."""
        if self.local.sessions:
            _, spoof_addrs = self.local.sessions[-1]
            count = spoof_addrs.get(addr, 0) + 1
            if count == 0:
                del spoof_addrs[addr]
            else:
                spoof_addrs[addr] = count
        self._spoof_addr_add(addr)
    def _spoof_addr_remove(self, addr):
        count = None
        with self.spoof_addrs_lock:
            if addr in self.spoof_addrs:
                count = self.spoof_addrs[addr] - 1
                if count == 0:
                    del self.spoof_addrs[addr]
                else:
                    self.spoof_addrs[addr] = count
        return count
    def spoof_addr_remove(self, addr):
        """Remove an address from being spoofed."""
        if self._spoof_addr_remove(addr) is not None and self.local.sessions:
            _, spoof_addrs = self.local.sessions[-1]
            count = spoof_addrs.get(addr, 0) - 1
            if count == 0:
                del spoof_addrs[addr]
            else:
                spoof_addrs[addr] = count
    def clear(self):
        queue = self.queues[threading.current_thread()]
        queue.clear()

def start_sniffing(target, *args, **kwargs):
    """A convenience function for sniffing that safely handles
    KeyboardInterrupt exceptions.
    """
    _assert_main_thread()
    old_handler = signal.getsignal(signal.SIGINT)
    assert(old_handler == signal.default_int_handler)
    interrupted = {}
    def interrupt_ignore(signum, frame):
        interrupted['signum'] = signum
        interrupted['frame'] = frame
        signal.signal(signal.SIGINT, signal.SIG_IGN)
    def interrupt_raise(signum, frame):
        interrupt_ignore(signum, frame)
        raise KeyboardInterrupt()
    sniffer = None
    try:
        signal.signal(signal.SIGINT, interrupt_ignore)
        target_args = kwargs.pop('args', ())
        target_kwargs = kwargs.pop('kwargs', {})
        kwargs['subscribe'] = True
        sniffer = Sniffer(*args, **kwargs)
        signal.signal(signal.SIGINT, interrupt_raise)
        if interrupted:
            raise KeyboardInterrupt()
        result = target(sniffer, *target_args, **target_kwargs)
        signal.signal(signal.SIGINT, interrupt_ignore)
        return result
    except:
        signal.signal(signal.SIGINT, interrupt_ignore)
        raise
    finally:
        _try_f(lambda: sniffer.shutdown())
        _try_f(lambda: signal.signal(signal.SIGINT, old_handler))
        if interrupted:
            raise KeyboardInterrupt()
